@@ -25,7 +25,7 @@ namespace ZeroTier {
 bool Path::send(const RuntimeEnvironment *RR,const void *data,unsigned int len,uint64_t now)
 {
 	if (RR->node->putPacket(_localAddress,address(),data,len)) {
-		sent(now);
+		_lastOut = now;
 		return true;
 	}
 	return false;
